#!/usr/bin/env python

"""
Annotate program node structures. The code in this module operates upon nodes
which are produced when simplifying AST node trees originating from the compiler
module.

Copyright (C) 2006, 2007 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

To use this module, the easiest approach is to use the annotate function which
provides support for annotating modules already processed by simplify and
fixnames:

annotate(builtins, None, importer)
annotate(module, builtins, importer)

At the most basic level, the most intricate approach involves obtaining an
Annotator object:

annotator = Annotator(importer)

Then, processing an existing module with it:

annotator.process(module)

If a module containing built-in classes and functions has already been
annotated, such a module should be passed in as an additional argument:

annotator.process(module, builtins)
"""

from simplify.simplified import *
import compiler
import os

class System:

    """
    A class maintaining the state of the annotation system. When the system
    counter can no longer be incremented by any annotation operation, the
    system may be considered stable and fully annotated.
    """

    def __init__(self):
        self.count = 0

    def init(self, node, attr="types"):

        "Initialise a 'node' for annotation."

        if not hasattr(node, attr):
            setattr(node, attr, set())

    def annotate(self, node, types, attr="types"):

        "Annotate the given 'node' with the given 'types'."

        self.init(node, attr)
        self.combine(getattr(node, attr), types)
        node.annotated = 1

    def combine(self, target, types):

        """
        Combine the 'target' list with the given 'types', counting new members.
        """

        for type in types:
            if type not in target:
                target.add(type)
                self.count += 1

system = System()

# Exceptions.

class AnnotationError(SimplifiedError):

    "An error in the annotation process."

    pass

class AnnotationMessage(Exception):

    "A lesser annotation error."

    pass

# Annotation.

class Annotator(Visitor):

    """
    The type annotator which traverses the program nodes, typically depth-first,
    and maintains a record of the current set of types applying to the currently
    considered operation. Such types are also recorded on the nodes, and a
    special "system" record is maintained to monitor the level of annotation
    activity with a view to recognising when no more annotations are possible.

    Throughout the annotation activity, type information consists of lists of
    Attribute objects where such objects retain information about the context of
    the type (since a value in the program may be associated with an object or
    class) and the actual type of the value being manipulated. Upon accessing
    attribute information on namespaces, additional accessor information is also
    exchanged - this provides a means of distinguishing between the different
    types possible when the means of constructing the namespace may depend on
    run-time behaviour.

    Covered: Assign, CheckType, Conditional, Global, Import, InvokeRef,
             InvokeFunction, LoadAttr, LoadExc, LoadName, LoadRef, LoadTemp,
             Module, Not, Pass, Raise, ReleaseTemp, ReturnFromBlock,
             ReturnFromFunction, StoreAttr, StoreName, StoreTemp, Subprogram,
             Try.
    """

    def __init__(self, importer):

        "Initialise the visitor with an 'importer'."

        Visitor.__init__(self)
        self.system = system
        self.importer = importer

        # Satisfy visitor issues.

        self.visitor = self

    def process(self, module, builtins=None):

        """
        Process the given 'module', using the optional 'builtins' to access
        built-in classes and functions.
        """

        self.current_subprograms = set()
        self.current_namespaces = []
        self.rerun_subprograms = {}
        self.namespace = None
        self.module = module

        # Process the module, supplying builtins if possible.

        self.builtins = builtins
        self.global_namespace = Namespace()

        if builtins is not None:
            self.builtins_namespace = builtins.namespace
        else:
            self.builtins_namespace = self.global_namespace

        # NOTE: Not declaring module namespace usage, even though it is used.

        self.process_node(module, self.global_namespace, 0)

    def process_node(self, node, locals, using_module_namespace):

        """
        Process a subprogram or module 'node', indicating the initial 'locals'.
        Note that this method may mutate nodes in the original program.
        """

        # Recursion test.

        if node in self.current_subprograms:
            if not self.rerun_subprograms.has_key(node):
                self.rerun_subprograms[node] = []
            self.rerun_subprograms[node].append(locals)
            return

        # Record the current subprogram and namespace.

        self.current_subprograms.add(node)

        # Determine the namespace.

        self.current_namespaces.append(self.namespace)
        self.namespace = locals

        # Add namespace details to any structure involved.

        if node.structure is not None:
            node.structure.namespace = Namespace()

            # Initialise bases where appropriate.

            if hasattr(node.structure, "bases"):
                base_refs = []
                for base in node.structure.bases:
                    self.dispatch(base)
                    base_refs.append(self.namespace.types)
                node.structure.base_refs = base_refs

        # Dispatch to the code itself.

        node.namespace = self.namespace
        self.set_module_namespace(using_module_namespace)

        self.dispatch(node)
        self.extract_results(node)

        # We may need to "re-run" subprograms because they were called
        # recursively with potentially different arguments. These argument types
        # are now tried out here.

        while self.rerun_subprograms.has_key(node):
            all_rerun_locals = self.rerun_subprograms[node]
            del self.rerun_subprograms[node]
            for rerun_locals in all_rerun_locals:
                #print "Re-running", node, "with", rerun_locals

                self.namespace = rerun_locals
                node.namespace = rerun_locals
                self.set_module_namespace(using_module_namespace)

                self.dispatch(node)
                self.extract_results(node)

        # Restore the previous subprogram and namespace.

        self.namespace = self.current_namespaces.pop()
        self.current_subprograms.remove(node)
        self.reset_module_namespace(using_module_namespace)

    def set_module_namespace(self, using_module_namespace):

        """
        In order to keep global accesses working, the module namespace must be
        adjusted.
        """

        if using_module_namespace:
            self.module.namespace = self.namespace

    def reset_module_namespace(self, using_module_namespace):

        """
        In order to keep global accesses working, the module namespace must be
        reset.
        """

        if using_module_namespace:
            self.module.namespace = self.namespace

    def extract_results(self, node):

        "Extract results from the namespace."

        node.namespace = self.namespace
        self.system.annotate(node, self.namespace.raises, "raises")
        self.system.annotate(node, self.namespace.returns, "returns")
        if hasattr(node, "return_locals"):
            node.return_locals.update(self.namespace.return_locals)

    def annotate(self, node, types=None):

        """
        Annotate the given 'node' in the system, using either the optional
        'types' or the namespace's current type information.
        """

        if types is None:
            self.system.annotate(node, self.namespace.types)
        else:
            self.system.annotate(node, types)

    def annotate_parameters(self, node, items):

        """
        Annotate the given 'node' using the given 'items' and updating the
        system's annotation counter.
        """

        for param, types in items:
            if not node.paramtypes.has_key(param):
                node.paramtypes[param] = set()
            self.system.combine(node.paramtypes[param], types)

    # Visitor methods.

    def default(self, node):

        """
        Process the given 'node', given that it does not have a specific
        handler.
        """

        raise AnnotationMessage, "Node '%s' not supported." % node

    def dispatch(self, node, *args):
        try:
            Visitor.dispatch(self, node, *args)
        except AnnotationError, exc:
            exc.add(node)
            raise
        except AnnotationMessage, exc:
            raise AnnotationError(exc, node)

    # Specific node methods.

    def visitAssign(self, assign):

        """
        Process the 'assign' node and its contents.
        """

        self.dispatches(assign.code)

    def visitCheckType(self, checktype):

        """
        Process the 'checktype' node, finding the possible types of the
        exception, and processing each choice to build a list of checked types
        for the exception.
        """

        inverted = getattr(checktype, "inverted", 0)
        self.dispatch(checktype.expr)

        expr_types = self.namespace.types
        choice_types = set()
        choices = []

        for choice in checktype.choices:
            choices.append(self.dispatch(choice))
            choice_types.update(self.namespace.types)

        for expr_type in expr_types:
            in_choices = expr_type.type.get_class() in choice_types

            # Filter out types not in the choices list unless the operation is
            # inverted; in which case, filter out types in the choices list.

            if not inverted and not in_choices or inverted and in_choices:
                self._prune_non_accesses(checktype.expr, expr_type)

    def visitConditional(self, conditional):

        """
        Process the 'conditional' node, processing the test, body and else
        clauses and recording their processed forms. The body and else clauses
        are processed within their own namespaces, and the test is also
        processed in its own namespace if 'isolate_test' is set on the
        'conditional' node.
        """

        # Conditionals keep local namespace changes isolated.
        # With Return nodes inside the body/else sections, the changes are
        # communicated to the caller.

        is_module = self.namespace is self.module.namespace

        # Where the test is closely associated with the body, save the namespace
        # before entering the test.

        if conditional.isolate_test:
            saved_namespace = self.namespace
            self.namespace = Namespace()
            if is_module:
                self.module.namespace = self.namespace
            self.namespace.merge_namespace(saved_namespace)

        self.dispatch(conditional.test)

        # Where the test may affect the body and the else clause, save the
        # namespace after processing the test.

        if not conditional.isolate_test:
            saved_namespace = self.namespace
            self.namespace = Namespace()
            if is_module:
                self.module.namespace = self.namespace
            self.namespace.merge_namespace(saved_namespace)

        # NOTE: Exception recording.

        else:
            test_raises = set()
            test_raises.update(self.namespace.raises)

        # Process the body clause.

        self.dispatches(conditional.body)
        body_namespace = self.namespace

        # Use the saved namespace as a template for the else clause.

        self.namespace = Namespace()
        if is_module:
            self.module.namespace = self.namespace
        self.namespace.merge_namespace(saved_namespace)

        # Process the else clause.

        self.dispatches(conditional.else_)
        else_namespace = self.namespace

        # Merge the body and else namespaces.

        self.namespace = Namespace()
        if is_module:
            self.module.namespace = self.namespace
        self.namespace.merge_namespace(body_namespace)
        self.namespace.merge_namespace(else_namespace)

        # NOTE: Test of exception type pruning based on the test/body.
        # Note that the checked exceptions are tested for re-raising.

        if conditional.isolate_test:
            for exc_type in test_raises:
                if exc_type not in body_namespace.raises:
                    self.namespace.revoke_exception_type(exc_type)

    def visitGlobal(self, global_):

        """
        Leave the 'global_' node unprocessed since namespaces should have
        already been altered to take global names into consideration.
        """

        pass

    def visitImport(self, import_):

        """
        Process the 'import_' node, importing the module with the stated name
        and storing details on the node.
        """

        module = self.importer.load(import_.name, self.builtins, import_.alias)
        if module is not None:
            self.namespace.set_types(set([module]))
        else:
            self.namespace.set_types(set())
        self.annotate(import_) # mainly for viewing purposes

    def _visitInvoke(self, invoke, invocation_types, have_args):

        """
        Process the 'invoke' node, using the given 'invocation_types' as the
        list of callables to be investigated for instantiation or for the
        invocation of functions or blocks. If 'have_args' is a true value, any
        invocation or instantiation will involve arguments.
        """

        # Now locate and invoke the subprogram. This can be complicated because
        # the target may be a class or object, and there may be many different
        # related subprograms.

        invocations = set()

        # Visit each callable in turn, finding subprograms.

        for attr in invocation_types:

            # Deal with class invocations by providing instance objects.
            # Here, each class is queried for the __init__ method, which may
            # exist for some combinations of classes in a hierarchy but not for
            # others.

            if isinstance(attr.type, GeneralClass):
                attributes = get_attributes(attr.type, "__init__")

            # Deal with object invocations by using __call__ methods.

            elif isinstance(attr.type, Instance):
                attributes = get_attributes(attr.type, "__call__")

            # Normal functions or methods are more straightforward.
            # Here, we model them using an attribute with no context and with
            # no associated accessor.

            else:
                attributes = [(attr, None)]

            # Inspect each attribute and extract the subprogram.

            for attribute, accessor in attributes:

                # If a class is involved, presume that it must create a new
                # object.

                if isinstance(attr.type, GeneralClass):

                    # Instantiate the class.

                    instance = self.new_instance(invoke, attr.type)

                    # For instantiations, switch the context.

                    if attribute is not None:
                        attribute = Attribute(instance, attribute.type)

                        # Request an instance-specific initialiser.

                        attribute = attr.type.get_attribute_for_instance(attribute, instance)

                # Skip cases where no callable is found.

                if attribute is not None:

                    # If a subprogram is defined, invoke it.

                    self.invoke_subprogram(invoke, attribute)
                    invocations.add(attribute.type)

                elif not isinstance(attr.type, GeneralClass):
                    print "Invocation type is None for", accessor

                else:

                    # Test to see if no arguments were supplied in cases where no
                    # initialiser was found.

                    if have_args:
                        raise AnnotationMessage, "No initialiser found for '%s' with arguments." % attr.type

            # Special case: initialisation.

            if isinstance(attr.type, GeneralClass):

                # Associate the instance with the result of this invocation.

                self.namespace.set_types(set([Attribute(None, instance)]))
                self.annotate(invoke)

        # Remember the invocations that were found, along with the return type
        # information.

        invoke.invocations = invocations
        self.namespace.set_types(getattr(invoke, "types", set()))

    def visitInvokeRef(self, invoke):

        """
        Process the 'invoke' node, first finding the callables indicated by the
        reference.
        """

        # Where the invocation belongs to an instance but the invoked subprogram
        # does not, request a special copy.

        instance = getattr(invoke, "instance", None)
        if instance is not None and getattr(invoke.ref, "instance", None) is None:
            if invoke.ref.copies.has_key(instance):
                invoke.ref = invoke.ref.copies[instance]
            else:
                invoke.ref = invoke.ref.copy(instance)
                #print "Created", invoke.ref, "for", getattr(invoke.ref, "instance", None)
                invoke.ref.module.simplifier.subnames[invoke.ref.full_name()] = invoke.ref
        invocation_types = [Attribute(None, invoke.ref)]
        self._visitInvoke(invoke, invocation_types, have_args=0)

    def visitInvokeFunction(self, invoke):

        """
        Process the 'invoke' node, first finding the callables indicated by the
        expression.
        """

        self.dispatch(invoke.expr)
        invocation_types = self.namespace.types

        # Invocation processing starts with making sure that the arguments have
        # been processed.

        self._visitInvoke(invoke, invocation_types, have_args=self.process_args(invoke))

    def visitLoadAttr(self, loadattr):

        """
        Process the 'loadattr' node, processing and storing the expression, and
        using the expression's types to construct records of accesses and
        non-accesses using the stated attribute name.
        """

        self.dispatch(loadattr.expr)
        types = set()
        raises = set()
        non_accesses = set()
        accesses = {}

        # For each expression type...

        for attr in self.namespace.types:

            # Find types for the named attribute.

            attributes = get_attributes(attr.type, loadattr.name)

            # Where no attributes exist...

            if not attributes:

                # Register new invalid accesses and mark a possible exception.

                if not attr in non_accesses:
                    non_accesses.add(attr)
                    exc = self.get_builtin_instances(loadattr, "AttributeError")
                    raises.update(exc)
                    self.namespace.raises.update(exc)

                    # Revoke this type from any name involved.

                    self._prune_non_accesses(loadattr.expr, attr)

            # For each type found...

            for attribute, accessor in attributes:

                # For actual attributes, register the type and remember the
                # access.

                if attribute is not None:
                    types.add(attribute)
                    if not accesses.has_key(attr.type):
                        accesses[attr.type] = []
                    if not (attribute, accessor) in accesses[attr.type]:
                        accesses[attr.type].append((attribute, accessor))

                # Otherwise, register new invalid accesses and note a possible
                # exception.

                else:
                    if not attr in non_accesses:
                        non_accesses.add(attr)
                        exc = self.get_builtin_instances(loadattr, "AttributeError")
                        raises.update(exc)
                        self.namespace.raises.update(exc)

                        # Revoke this type from any name involved.

                        self._prune_non_accesses(loadattr.expr, attr)

        if not types:
            print "No attribute found for", loadattr.name, "given", self.namespace.types

        # Remember the result types.

        self.namespace.set_types(types)
        loadattr.non_accesses = non_accesses
        loadattr.accesses = accesses
        loadattr.raises = raises
        self.annotate(loadattr)

    def _prune_non_accesses(self, expr, attr):

        """
        Prune type information from 'expr' where the given 'attr' has been
        shown to be a non-access.
        """

        if isinstance(expr, LoadName):
            self.namespace.revoke(expr.name, attr)
        elif isinstance(expr, LoadExc):
            self.namespace.revoke_exception_type(attr)
        elif isinstance(expr, LoadTemp):
            self.namespace.revoke_temp_type(getattr(expr, "index", None), attr)

        # LoadAttr cannot be pruned since this might unintentionally prune
        # legitimate types from other applications of the referenced type, it
        # almost certainly doesn't take "concurrent" mutation into
        # consideration (where in a running program, the pruned type is actually
        # reintroduced, making the pruning invalid), and there is no easy way of
        # preserving the meaning of a namespace without either creating lots of
        # specialised instances, and even then...

        #elif isinstance(expr, LoadAttr):
        #    for expr_attr in expr.expr.types:
        #        if hasattr(expr_attr.type, "namespace"):
        #            expr_attr.type.namespace.revoke(expr.name, attr)

    def visitLoadExc(self, loadexc):

        """
        Process the 'loadexc' node, discovering the possible exception types
        raised.
        """

        self.namespace.set_types(self.namespace.raises)
        self.annotate(loadexc)

    def visitLoadName(self, loadname):

        """
        Process the 'loadname' node, processing the name information on the node
        to determine which types are involved with the name.
        """

        self.namespace.set_types(self.namespace.load(loadname.name))
        self.annotate(loadname)

    def visitLoadRef(self, loadref):

        """
        Process the 'loadref' node, obtaining type information about the
        reference stated on the node.
        """

        self.namespace.set_types(set([Attribute(None, loadref.ref)]))
        self.annotate(loadref)

    def visitLoadTemp(self, loadtemp):

        """
        Process the 'loadtemp' node, obtaining type information about the
        temporary variable accessed, and removing variable information where the
        'release' attribute has been set on the node.
        """

        index = getattr(loadtemp, "index", None)
        try:
            if getattr(loadtemp, "release", 0):
                self.namespace.set_types(self.namespace.temp[index].pop())
            else:
                self.namespace.set_types(self.namespace.temp[index][-1])
        except KeyError:
            raise AnnotationMessage, "Temporary store index '%s' not defined." % index
        self.annotate(loadtemp)

    def visitMakeTuple(self, maketuple):

        """
        Process the 'maketuple' node and its contents.
        """

        # Get a tuple and populate it with type information for the contents.

        tuples = self.get_builtin_instances(maketuple, "tuple")

        # NOTE: This is dependent on the tuple definition in the builtins.

        for i, node in enumerate(maketuple.nodes):
            self.dispatch(node)
            for t in tuples:
                t.type.namespace.add("value", self.namespace.types)
                t.type.namespace.add("__value__%d" % i, self.namespace.types)

        self.namespace.set_types(tuples)
        self.annotate(maketuple)

    def visitModule(self, module):

        """
        Process the 'module' and its contents.
        """

        self.dispatches(module.code)

    def visitNot(self, not_):

        "Process the 'not_' node and its expression."

        self.dispatch(not_.expr)

    def visitPass(self, pass_):

        "Leave the 'pass_' node unprocessed."

        pass

    def visitRaise(self, raise_):

        """
        Process the 'raise_' node, processing any traceback information along
        with the raised exception expression, converting the node into a kind of
        invocation where the expression is found not to be an invocation itself.
        This node affects the namespace, adding exception types to the list of
        those raised in the namespace.
        """

        if raise_.traceback is not None:
            self.dispatch(raise_.traceback)
        self.dispatch(raise_.expr)

        # Handle bare name exceptions by converting any classes to instances.

        if not isinstance(raise_.expr, InvokeFunction):
            raise_.pos_args = []
            raise_.kw_args = {}
            raise_.star = None
            raise_.dstar = None
            types = set()
            for attr in self.namespace.types:
                if isinstance(attr.type, GeneralClass):
                    self._visitInvoke(raise_, [attr], have_args=0)
                types.update(self.namespace.types)
        else:
            types = self.namespace.types
            
        self.namespace.raises.update(types)

    def visitReleaseTemp(self, releasetemp):

        """
        Process the 'releasetemp' node, removing temporary variable information
        from the current namespace.
        """

        index = getattr(releasetemp, "index", None)
        try:
            self.namespace.temp[index].pop()
        except KeyError:
            raise AnnotationMessage, "Temporary store index '%s' not defined." % index
        except IndexError:
            pass #raise AnnotationMessage, "Temporary store index '%s' is empty." % index

    def visitResetExc(self, resetexc):
        self.namespace.raises = set()

    def visitReturn(self, return_):

        """
        Process the 'return_' node, processing any expression and obtaining type
        information to be accumulated in the current namespace's list of return
        types. A snapshot of the namespace is taken for the purposes of
        reconciling or merging namespaces where subprograms actually share
        locals with their callers.
        """

        if hasattr(return_, "expr"):
            self.dispatch(return_.expr)
            self.namespace.returns.update(self.namespace.types)
            self.annotate(return_)
        self.namespace.snapshot()

    visitReturnFromBlock = visitReturn
    visitReturnFromFunction = visitReturn

    def visitStoreAttr(self, storeattr):

        """
        Process the 'storeattr' node, processing the expression and target, and
        using the type information obtained to build records of legitimate
        writes to the stated attribute, along with "impossible" non-writes to
        the attribute.
        """

        self.dispatch(storeattr.expr)
        expr = self.namespace.types
        self.dispatch(storeattr.lvalue)
        writes = {}
        non_writes = set()
        for attr in self.namespace.types:
            # NOTE: Impose "atomic" constraints on certain types.
            if attr is None:
                if not attr in non_writes:
                    non_writes.add(attr)
                continue
            attr.type.namespace.add(storeattr.name, expr)
            writes[attr.type] = attr.type.namespace.load(storeattr.name)
        if not writes:
            print "Unable to store attribute", storeattr.name, "given", self.namespace.types
        storeattr.writes = writes
        storeattr.non_writes = non_writes

    def visitStoreName(self, storename):

        """
        Process the 'storename' node, processing the expression on the node and
        associating the type information obtained with the stated name in the
        current namespace.
        """

        self.dispatch(storename.expr)
        self.namespace.store(storename.name, self.namespace.types)
        self.annotate(storename)

    def visitStoreTemp(self, storetemp):

        """
        Process the 'storetemp' node, processing the expression on the node and
        associating the type information obtained with a temporary variable in
        the current namespace.
        """

        self.dispatch(storetemp.expr)
        index = getattr(storetemp, "index", None)
        if not self.namespace.temp.has_key(index):
            self.namespace.temp[index] = []
        self.namespace.temp[index].append(self.namespace.types)

    def visitSubprogram(self, subprogram):

        """
        Process the 'subprogram' node, processing its contents (a group of nodes
        comprising the subprogram).
        """

        self.dispatches(subprogram.code)

    def visitTry(self, try_):

        """
        Process the 'try_' node, processing the body clause in its own namespace
        derived from the current namespace, processing any handler clause using
        the namespace information accumulated in the body, and processing any
        else and finally clauses, attempting to supply each with appropriate
        namespace information.
        """

        is_module = self.namespace is self.module.namespace

        self.dispatches(try_.body)

        # Save the namespace from the body.

        body_namespace = Namespace()
        body_namespace.merge_namespace(self.namespace)

        # Process the handler.

        if hasattr(try_, "handler"):
            self.dispatches(try_.handler)

        # Save the namespace from the handler.

        handler_namespace = Namespace()
        handler_namespace.merge_namespace(self.namespace)

        # Remember the raised exceptions encountered so far.

        raises = self.namespace.raises

        # Process the else clause.

        if hasattr(try_, "else_"):

            # Restore the body namespace for the else clause.

            self.namespace = body_namespace
            if is_module:
                self.module.namespace = self.namespace

            # Empty the raised exceptions for the else clause.

            self.namespace.raises = set()
            self.dispatches(try_.else_)
            self.namespace.raises = raises

        # Merge the namespaces.

        self.namespace = Namespace()
        if is_module:
            self.module.namespace = self.namespace
        self.namespace.merge_namespace(body_namespace)
        self.namespace.merge_namespace(handler_namespace)

        # Process the finally clause, if any.

        self.dispatches(try_.finally_)

    def visitYield(self, yield_):
        raise NotImplementedError, "The yield statement is not currently supported."

    # Utility methods.

    def get_builtin_instances(self, node, name):
        return set([Attribute(None, self.new_instance(node, attr.type)) for attr in self.builtins_namespace[name]])

    def new_instance(self, node, type):

        "For the given 'node', obtain an instance from the given 'type'."

        if not type.has_instance(node):
            instance = Instance()
            instance.namespace = Namespace()
            instance.namespace.store("__class__", set([Attribute(None, type)]))
            type.add_instance(node, instance)
        else:
            instance = type.get_instance(node)

        return instance

    def invoke_subprogram(self, invoke, attribute):

        """
        Invoke using the given 'invoke' node the subprogram represented by the
        given 'attribute'.
        """

        # Test for context information, making it into a real attribute.

        if attribute.context is not None:
            context = Attribute(None, attribute.context)
            target = attribute.type
        else:
            context = None
            target = attribute.type

        # Test to see if anything has changed.

        if hasattr(invoke, "syscount") and invoke.syscount.has_key(target) and invoke.syscount[target] == self.system.count:
            return

        # Remember the state of the system.

        else:
            if not hasattr(invoke, "syscount"):
                invoke.syscount = {}
            invoke.syscount[target] = self.system.count

        # Provide the correct namespace for the invocation.
        # This may be a "shared" namespace...

        if invoke.share_locals:
            namespace = Namespace()
            namespace.merge_namespace(self.namespace, everything=0)
            using_module_namespace = self.namespace is self.module.namespace

        # Or it may be a structure...

        elif target.structure:
            namespace = Namespace()
            using_module_namespace = 0

        # Or it may be a new namespace populated with the supplied parameters.

        else:
            items = self.make_items(invoke, target, context)
            namespace = Namespace()
            namespace.merge_items(items)
            using_module_namespace = 0

            # NOTE: Avoid PEP 227 (nested scopes) whilst permitting references to a
            # NOTE: subprogram within itself. Do not define the name of the function
            # NOTE: within a method definition.

            if target.name is not None and not target.is_method:
                namespace.store(target.name, set([Attribute(None, target)]))

        # Process the subprogram.

        self.process_node(target, namespace, using_module_namespace)

        # NOTE: Improve and verify this.
        # If the invocation returns a value, acquire the return types.

        if target.returns_value:
            self.namespace.set_types(target.returns)
            self.annotate(invoke)

        # If it is a normal block, merge the locals.
        # This can happen in addition to the above because for things like
        # logical expressions, the namespace can be modified whilst values are
        # returned as results.

        if invoke.share_locals:
            self.namespace.reset()

            # Merge the locals snapshots.

            for locals in target.return_locals:

                # For blocks returning values (such as operations), do not merge
                # snapshots or results.

                if target.returns_value:
                    self.namespace.merge_namespace(locals, everything=0, temp=0)

                # For blocks not returning values (such as loops), merge
                # snapshots and results since they contain details of genuine
                # returns.

                else:
                    self.namespace.merge_namespace(locals, temp=0)

        # Incorporate any raised exceptions.

        invoke.raises.update(target.raises)
        self.namespace.raises.update(target.raises)

    def process_args(self, invocation):

        """
        Process the arguments associated with an 'invocation'. Return whether
        any arguments were processed.
        """

        self.dispatches(invocation.pos_args)
        self.dispatch_dict(invocation.kw_args)

        # Get type information for star and dstar arguments.

        if invocation.star is not None:
            param, default = invocation.star
            self.dispatch(default)
            invocation.star = param, default

        if invocation.dstar is not None:
            param, default = invocation.dstar
            self.dispatch(default)
            invocation.dstar = param, default

        if invocation.pos_args or invocation.kw_args or invocation.star or invocation.dstar:
            return 1
        else:
            return 0

    def make_items(self, invocation, subprogram, context):

        """
        Make an items mapping for the 'invocation' of the 'subprogram' using the
        given 'context' (which may be None).
        """

        # NOTE: Support class methods!

        if context is not None and isinstance(context.type, Instance):
            pos_args = [Self(context)] + invocation.pos_args
        else:
            pos_args = invocation.pos_args

        # Duplicate the keyword arguments - we remove them in processing below.

        kw_args = {}
        kw_args.update(invocation.kw_args)

        # Sort the arguments into positional and keyword arguments.

        params = subprogram.params
        items = []
        star_args = []
        consumed_args = []

        # Match each positional argument, taking excess arguments as star args.

        for arg in pos_args:
            if params:
                param, default = params[0]
                if hasattr(arg, "types"):
                    items.append((param, arg.types))
                else:
                    items.append((param, set())) # Annotation has not succeeded.

                # Record consumption order.

                consumed_args.append(arg)
                params = params[1:]
            else:
                star_args.append(arg)

        # Collect the remaining defaults.

        while params:
            param, default = params[0]
            if kw_args.has_key(param):
                arg = kw_args[param]
                del kw_args[param]
            elif default is not None:
                self.dispatch(default)
                arg = default
            else:
                raise AnnotationMessage, "No argument supplied in '%s' for parameter '%s'." % (subprogram, param)
            if hasattr(arg, "types"):
                items.append((param, arg.types))
            else:
                items.append((param, set())) # Annotation has not succeeded.

            # Record consumption order (this is not the Keyword node, however).

            consumed_args.append(arg)
            params = params[1:]

        dstar_args = kw_args.items()

        # Construct temporary objects.

        if star_args:
            star_invocation = self.make_star_args(invocation, subprogram, star_args)
            self.dispatch(star_invocation)
            star_types = star_invocation.types
        else:
            star_types = None

        if dstar_args:
            dstar_invocation = self.make_dstar_args(invocation, subprogram, dstar_args)
            self.dispatch(dstar_invocation)
            dstar_types = dstar_invocation.types
        else:
            dstar_types = None

        # NOTE: Merge the objects properly.

        star_types = star_types or invocation.star and invocation.star.types
        dstar_types = dstar_types or invocation.dstar and invocation.dstar.types

        # Add star and dstar.

        if star_types is not None:
            if subprogram.star is not None:
                param, default = subprogram.star
                items.append((param, star_types))
            else:
                raise AnnotationMessage, "Invocation provides unwanted *args."
        elif subprogram.star is not None:
            param, default = subprogram.star
            if not hasattr(default, "types") or not default.types:
                self.dispatch(default) # NOTE: Review reprocessing.
            items.append((param, default.types))

        if dstar_types is not None:
            if subprogram.dstar is not None:
                param, default = subprogram.dstar
                items.append((param, dstar_types))
            else:
                raise AnnotationMessage, "Invocation provides unwanted **args."
        elif subprogram.dstar is not None:
            param, default = subprogram.dstar
            if not hasattr(default, "types") or not default.types:
                self.dispatch(default) # NOTE: Review reprocessing.
            items.append((param, default.types))

        # Record the parameter types.

        self.annotate_parameters(subprogram, items)
        invocation.consumed_args[subprogram] = consumed_args
        return subprogram.paramtypes.items()

    def make_star_args(self, invocation, subprogram, star_args):

        "Make a subprogram which initialises a list containing 'star_args'."

        if not hasattr(invocation, "stars"):
            invocation.stars = {}

        if not invocation.stars.has_key(subprogram.full_name()):
            instance = getattr(invocation, "instance", None)

            code = [
                Return(
                    instance=instance,
                    expr=MakeTuple(
                        instance=instance,
                        nodes=star_args
                        )
                    )
                ]

            new_subprogram = Subprogram(
                instance=instance,
                name=None,
                returns_value=1,
                params=[],
                star=None,
                dstar=None,
                code=code
                )

            subprogram.module.simplifier.subnames[new_subprogram.full_name()] = new_subprogram

            invocation.stars[subprogram.full_name()] = InvokeRef(
                invocation.original,
                instance=instance,
                produces_result=1,
                ref=new_subprogram
                )

        return invocation.stars[subprogram.full_name()]

    def make_dstar_args(self, invocation, subprogram, dstar_args):

        """
        Make a subprogram which initialises a dictionary built from the given
        'dstar_args'.
        """

        if not hasattr(invocation, "dstars"):
            invocation.dstars = {}

        if not invocation.dstars.has_key(subprogram.full_name()):
            instance = getattr(invocation, "instance", None)

            code=[
                StoreTemp(
                    instance=instance,
                    expr=InvokeFunction(
                        invocation.original,
                        instance=instance,
                        expr=LoadAttr(
                            instance=instance,
                            expr=LoadRef(
                                instance=instance,
                                ref=self.builtins
                                ),
                            name="dict",
                            nstype="module",
                            )
                        )
                    )
                ]

            for arg, value in dstar_args:

                # NOTE: Constant not added to table.

                constant = Constant(name=repr(arg), value=arg)
                code += [
                    StoreTemp(
                        instance=instance,
                        expr=InvokeFunction(
                            instance=instance,
                            expr=LoadName(
                                instance=instance,
                                name=constant.typename
                                )
                            ),
                        index="const"
                        ),
                    InvokeFunction(
                        invocation.original,
                        instance=instance,
                        expr=LoadAttr(
                            instance=instance,
                            expr=LoadTemp(
                                instance=instance
                                ),
                            name="__setitem__"
                            ),
                        args=[
                            LoadTemp(
                                instance=instance,
                                index="const",
                                release=1
                                ),
                            value
                            ]
                        )
                    ]

            code += [
                Return(
                    instance=instance,
                    expr=LoadTemp(
                        instance=instance,
                        release=1
                        )
                    )
                ]

            new_subprogram = Subprogram(
                instance=instance,
                name=None,
                returns_value=1,
                params=[],
                star=None,
                dstar=None,
                code=code
                )
            subprogram.module.simplifier.subnames[new_subprogram.full_name()] = new_subprogram

            invocation.dstars[subprogram.full_name()] = InvokeRef(
                invocation.original,
                instance=instance,
                produces_result=1,
                ref=new_subprogram
                )

        return invocation.dstars[subprogram.full_name()]

# Namespace-related abstractions.

def combine(target, additions):

    """
    Merge into the 'target' sequence the given 'additions', preventing duplicate
    items.
    """

    for addition in additions:
        if addition not in target:
            target.append(addition)

def find_attributes(structure, name):

    """
    Find for the given 'structure' all attributes for the given 'name', visiting
    base classes where appropriate and returning the attributes in order of
    descending precedence for all possible base classes.

    The elements in the result list are 2-tuples which contain the attribute and
    the structure involved in accessing the attribute.
    """

    # First attempt to search the instance/class namespace.

    try:
        l = structure.namespace.load(name)
        attributes = []
        for attribute in l:
            attributes.append((attribute, structure))

    # If that does not work, attempt to investigate any class or base classes.

    except KeyError:
        attributes = []

        # Investigate any instance's implementing class.

        if isinstance(structure, Instance):
            for attr in structure.namespace.load("__class__"):
                cls = attr.type
                l = get_attributes(cls, name)
                combine(attributes, l)

        # Investigate any class's base classes.

        elif isinstance(structure, GeneralClass):

            # If no base classes exist, return an indicator that no attribute
            # exists.

            if not structure.base_refs:
                return [(None, structure)]

            # Otherwise, find all possible base classes.

            for base_refs in structure.base_refs:
                base_attributes = []

                # For each base class, find attributes either in the base
                # class or its own base classes.

                for base_ref in base_refs:
                    l = get_attributes(base_ref, name)
                    combine(base_attributes, l)

                combine(attributes, base_attributes)

    return attributes

def get_attributes(structure, name):

    """
    Return all possible attributes for the given 'structure' having the given
    'name', wrapping each attribute in an Attribute object which includes
    context information for the attribute access.

    The elements in the result list are 2-tuples which contain the attribute and
    the structure involved in accessing the attribute.
    """

    if isinstance(structure, Attribute):
        structure = structure.type

    results = []
    for attribute, accessor in find_attributes(structure, name):

        # Detect class attribute access via instances.

        if attribute is not None and isinstance(structure, Instance) and isinstance(accessor, GeneralClass):
            attribute = accessor.get_attribute_for_instance(attribute, structure)

        # Produce an attribute with the appropriate context.
        # Modules should not affect the context.

        if attribute is not None and isinstance(structure, Module):
            results.append((attribute, accessor))

        # Access via classes and instances should set those as the context.

        elif attribute is not None and isinstance(structure, Structure):
            results.append((Attribute(structure, attribute.type), accessor))

        # Other accesses are returned unmodified.

        else:
            results.append((attribute, accessor))

    return results

def prompt(vars):
    try:
        while 1:
            s = raw_input("> ")
            print eval(s, vars)
    except EOFError:
        pass

# Convenience functions.

def annotate(module, builtins, importer):

    """
    Annotate the given 'module', also employing the 'builtins' module which may
    be specified as None. Use the 'importer' to find and load modules.
    """

    annotator = Annotator(importer)
    if builtins is not None:
        annotator.process(module, builtins)
    else:
        annotator.process(module)

# vim: tabstop=4 expandtab shiftwidth=4
